﻿using System;
using System.IO;
using System.Linq;
using System.Security.Cryptography;
using System.Threading;
using System.Threading.Tasks;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using Newtonsoft.Json;
using steam_dropler.Model;
using SteamKit2;
using SteamKit2.Authentication;
using SteamKit2.Discovery;
using SteamKit2.Internal;


namespace steam_dropler.Steam
{
    public class SteamLoginHandler
    {
        private readonly SteamClient _client;
        private readonly SteamUser _sUser;
        private readonly SteamFriends _sFriends;
        private bool IsConnectedAndLoggedOn => _client?.SteamID != null;

        private readonly SteamUnifiedMessages steamUnifiedMessages;
        private readonly SteamUnifiedMessages.UnifiedService<IFriendMessages> UnifiedFriendMessagesService;

        private int tryLoginCount;


        private readonly AccountConfig _steamAccount;

        private readonly TaskCompletionSource<EResult> _loginTcs;

        private ServerRecord _serverRecord;

        private int LogLineCnt;
        private string LogLineBuff;
        private string _refreshToken;

        public ServerRecord ServerRecord => _serverRecord;

        public SteamLoginHandler(AccountConfig steamAccount, SteamClient client, CallbackManager manager)
        {
            _steamAccount = steamAccount;
            _client = client;
            _sUser = _client.GetHandler<SteamUser>();
            _sFriends = _client.GetHandler<SteamFriends>();

            steamUnifiedMessages = _client.GetHandler<SteamUnifiedMessages>();
            UnifiedFriendMessagesService = steamUnifiedMessages.CreateService<IFriendMessages>();

            _loginTcs = new TaskCompletionSource<EResult>();

            manager.Subscribe<SteamClient.ConnectedCallback>(OnConnected);
            manager.Subscribe<SteamClient.DisconnectedCallback>(OnDisconnected);
            manager.Subscribe<SteamUser.LoggedOnCallback>(OnLoggedOn);
            manager.Subscribe<SteamUser.LoggedOffCallback>(OnLoggedOff);
            manager.Subscribe<SteamUser.AccountInfoCallback>(OnAccountInfo);
            manager.Subscribe<SteamFriends.FriendMsgCallback>(OnFriendMsgCallback);
        }

        public Task<EResult> Login(ServerRecord serverRecord)
        {
            _serverRecord = serverRecord;
            _client.Connect(_serverRecord);
            return _loginTcs.Task;
        }
        

        async void OnConnected(SteamClient.ConnectedCallback callback)
        {
            LogLineCnt++;
            Console.WriteLine($"{Util.hdl(_steamAccount.Alias)}Connected to Steam server ({_serverRecord.GetHost()}, port: {_serverRecord.GetPort()}), try logging in...");
            if (_steamAccount.AccessToken != null)
            {
                _refreshToken = _steamAccount.AccessToken;
                _sUser.LogOn(new SteamUser.LogOnDetails
                {
                    Username = _steamAccount.Name,
                    AccessToken = _steamAccount.AccessToken,
                    ShouldRememberPassword = true
                });
            } else {
               
                IAuthenticator auth;
                switch (_steamAccount.AuthType)
                {
                    case AuthType.Console:
                        auth = new UserConsoleAuthenticator();
                        break;
                   
                    case AuthType.WithSecretKey:
                        auth = new TwoFactorAuth(_steamAccount.SharedSecret ?? _steamAccount.MobileAuth.SharedSecret);
                        break;

                    case AuthType.Device:
                    default:
                        auth = new DeviceAuth();
                        break;

                }
                try
                {
                    var authSession = await _client.Authentication.BeginAuthSessionViaCredentialsAsync(new AuthSessionDetails
                    {
                        Username = _steamAccount.Name,
                        Password = _steamAccount.Password,
                        ClientOSType = EOSType.Windows10,
                        DeviceFriendlyName = _steamAccount.Name + "pc",
                        PlatformType = EAuthTokenPlatformType.k_EAuthTokenPlatformType_SteamClient,
                        IsPersistentSession = true,
                        WebsiteID = "Client",
                        Authenticator = auth,
                    });

                
                    var pollResponse = await authSession.PollingWaitForResultAsync();
                    _refreshToken = pollResponse.RefreshToken;
                    _sUser.LogOn(new SteamUser.LogOnDetails
                    {
                        Username = pollResponse.AccountName,
                        AccessToken = pollResponse.RefreshToken,
                        ShouldRememberPassword = true
                    });
                }
                catch (Exception e)
                {
                    _steamAccount.Save();
                    Console.WriteLine("Exception while logon to Steam: {0}", e.Message);
                    _loginTcs?.TrySetResult(EResult.UnexpectedError);
                }
            }
            
           
        }


        void OnDisconnected(SteamClient.DisconnectedCallback callback)
        {
            tryLoginCount++;
            string s = $"{Util.hdl(_steamAccount.Alias)}Disconnected({tryLoginCount}) from Steam";
            if (callback.UserInitiated) {Console.WriteLine(s + $" (UserInitiated)"); return;}
            if (LogLineCnt > 1) {LogLineBuff += $"{s}\n";} else Console.WriteLine(s);
            if (_steamAccount.IdleNow) {if (!String.IsNullOrEmpty(LogLineBuff)) {Console.Write(LogLineBuff); LogLineBuff = "";} _steamAccount.Action = "Serr"; return;}
            else if (LogLineCnt > 0 || tryLoginCount > 4) {
                if (_steamAccount.Action != "Ber1") {
                    _steamAccount.LastRun =  DateTime.UtcNow.AddMinutes(-(_steamAccount.TimeConfig.IdleTime + _steamAccount.TimeConfig.PauseBeatwinIdleTime - 1));
                } else {_steamAccount.IdleNow = true;}
                _steamAccount.Action = "none";
                _loginTcs?.SetResult(EResult.NoConnection);
                return;
            }
            if (_steamAccount.Action != "Ber1") {Thread.Sleep(TimeSpan.FromSeconds(5));} else Thread.Sleep(TimeSpan.FromSeconds(10));
            _client.Connect(_serverRecord);
        }

        void OnLoggedOn(SteamUser.LoggedOnCallback callback)
        {
            LogLineCnt--;
            if (!String.IsNullOrEmpty(LogLineBuff)) {Console.Write(LogLineBuff); LogLineBuff = "";}
            

            if (callback.Result != EResult.OK)
            {
                if (callback.Result == EResult.Expired)
                {
                    _steamAccount.AccessToken = null;
                    _steamAccount.Save();
                }
                else if (callback.Result == EResult.NoConnection)
                {
                    SteamServerList.SetBadServer(_serverRecord);
                    _serverRecord = SteamServerList.GetServerRecord();
                }
                Console.Write($"{Util.hdl(_steamAccount.Alias)}");
                Console.ForegroundColor = ConsoleColor.Red;
                Console.WriteLine("Unable to logon to Steam: {0} / {1}", callback.Result, callback.ExtendedResult);
                Console.ResetColor();
                return;
            }
            _steamAccount.SteamId = _client.SteamID;
            _steamAccount.AccessToken = _refreshToken;
            Console.Write($"{Util.hdl(_steamAccount.Alias)}");
            Console.Write("Successfully logged on! ");

            if (_steamAccount.ShowStatus == "Online") {
                //Console.Write($"{Util.hdl(_steamAccount.Alias)}");
                //Console.Write($"PersonaState: {_sFriends.GetPersonaState()} ");
                _sFriends.SetPersonaState(EPersonaState.Online);
                Console.Write("Bot ");
                Console.ForegroundColor = ConsoleColor.DarkGreen;
                Console.WriteLine("Online");
                Console.ResetColor();
            } else Console.WriteLine("");
            tryLoginCount = 0;
            _loginTcs?.SetResult(callback.Result);
        }

        void OnLoggedOff(SteamUser.LoggedOffCallback callback)
        {
            string s = $"{Util.hdl(_steamAccount.Alias)}Logged off of Steam: {callback.Result}";
            if (LogLineCnt > 1) {LogLineBuff += $"{s}\n";}
            else Console.WriteLine(s);
        }

        void OnAccountInfo( SteamUser.AccountInfoCallback callback )
        {
            //Console.Write($"[{_steamAccount.Alias}] PersonaState: ");
            //Console.WriteLine(_sFriends.GetPersonaState());
        }

        void OnFriendMsgCallback(SteamFriends.FriendMsgCallback callback)
        {
            if (callback.EntryType == EChatEntryType.ChatMsg && callback.Sender.ConvertToUInt64() == MainConfig.Config.OwnerID) {
                //Console.WriteLine($"{Util.hdl(_steamAccount.Alias)}InMsg:{callback.Message}");
                //_sFriends.SendChatMessage(callback.Sender, EChatEntryType.ChatMsg, "Ok");
                if (callback.Message[0] == '!') chkBots(callback.Sender, _steamAccount.Alias, callback.Message);
            }
        }


        void chkBots(SteamID Sender, string bot, string Action)
        {
            string[] Acr = Regex.Replace(Action, @"\s+", " ").Split(' '); int AcrLen = Acr.Length;
            string _act = Acr[0], _bots = "", _prm = "", rz = "/code \u2063\ufeff"; int dtm;
            if (AcrLen == 1) {_bots = bot;}
            else if (AcrLen == 2) {if (Acr[1][0] == '@') {_bots = Acr[1].Substring(1);} else {_bots = bot; _prm = Acr[1];}}
            else if (AcrLen > 2) {if (Acr[1][0] == '@') {_bots = Acr[1].Substring(1); _prm = Acr[2];} else {_bots = bot; _prm = Acr[1];}}
            string[] Abots = _bots.Split(',');
            if (_act == "!stat"){
                string rzo = "", rzf = "", rzd = "", dcfg, tcfg;
                if (Abots[0] == "all" || Abots[0] == "ion"){
                    var _acfg = Worker._accounts.Where(t => t.IdleEnable && t.IdleNow);
                    foreach (var _cfg in _acfg) {
                        dcfg = ""; foreach (var _dcfg in _cfg.DropConfig) {dcfg += $"{_dcfg.Item1}:{_dcfg.Item2};";}
                        tcfg = $"{_cfg.TimeConfig.IdleTime},{_cfg.TimeConfig.PauseBeatwinIdleTime}";
                        dtm  = _cfg.TimeConfig.IdleTime - (int)(DateTime.UtcNow - _cfg.LastRun.Value).TotalMinutes;
                        rzo += $"{Util.hdl(_cfg.Alias, 3)}Idle:'True'  DropCfg:{dcfg}  TimeCfg:{tcfg} (Стоп через {dtm} мин.)\n";
                    }
                }
                if (Abots[0] == "all" || Abots[0] == "iof"){
                    var _acfg = Worker._accounts.Where(t => t.IdleEnable && !t.IdleNow);
                    foreach (var _cfg in _acfg) {
                        dcfg = ""; foreach (var _dcfg in _cfg.DropConfig) {dcfg += $"{_dcfg.Item1}:{_dcfg.Item2};";}
                        tcfg = $"{_cfg.TimeConfig.IdleTime},{_cfg.TimeConfig.PauseBeatwinIdleTime}";
                        dtm  = (_cfg.TimeConfig.IdleTime + _cfg.TimeConfig.PauseBeatwinIdleTime) - (int)(DateTime.UtcNow - _cfg.LastRun.Value).TotalMinutes;
                        if (dtm > 0) {rzf += $"{Util.hdl(_cfg.Alias, 3)}Idle:False   DropCfg:{dcfg}  TimeCfg:{tcfg} (Старт через {dtm} мин.)\n";}
                        else rzf += $"{Util.hdl(_cfg.Alias, 3)}Idle:False   DropCfg:{dcfg}  TimeCfg:{tcfg} (В очереди на запуск)\n";
                    }
                }
                if (Abots[0] == "all" || Abots[0] == "ids"){
                    var _acfg = Worker._accounts.Where(t => !t.IdleEnable);
                    foreach (var _cfg in _acfg) {
                        dcfg = ""; foreach (var _dcfg in _cfg.DropConfig) {dcfg += $"{_dcfg.Item1}:{_dcfg.Item2};";}
                        tcfg = $"{_cfg.TimeConfig.IdleTime},{_cfg.TimeConfig.PauseBeatwinIdleTime}";
                        rzd += $"{Util.hdl(_cfg.Alias, 3)}False: Бот отключен   DropCfg:{dcfg}  TimeCfg:{tcfg}\n";
                    }
                }
                if (Abots[0] != "all" && Abots[0] != "ion" && Abots[0] != "iof" && Abots[0] != "ids"){
                    foreach (var _bot in Abots) {
                        var _acfg = Worker._accounts.FirstOrDefault(t =>t.Alias==_bot);
                        if (_acfg != null) {
                            dcfg = ""; foreach (var _dcfg in _acfg.DropConfig) {dcfg += $"{_dcfg.Item1}:{_dcfg.Item2};";}
                            tcfg = $"{_acfg.TimeConfig.IdleTime},{_acfg.TimeConfig.PauseBeatwinIdleTime}";
                            if (_acfg.IdleEnable) {
                                if (_acfg.IdleNow) {
                                    dtm  = _acfg.TimeConfig.IdleTime - (int)(DateTime.UtcNow - _acfg.LastRun.Value).TotalMinutes;
                                    rzo += $"{Util.hdl(_acfg.Alias, 3)}Idle:'True'  DropCfg:{dcfg}  TimeCfg:{tcfg} (Стоп через {dtm} мин.)\n";
                                } else {
                                    dtm  = (_acfg.TimeConfig.IdleTime + _acfg.TimeConfig.PauseBeatwinIdleTime) - (int)(DateTime.UtcNow - _acfg.LastRun.Value).TotalMinutes;
                                    if (dtm > 0) {rzo += $"{Util.hdl(_acfg.Alias, 3)}Idle:False   DropCfg:{dcfg}  TimeCfg:{tcfg} (Старт через {dtm} мин.)\n";}
                                    else rzo += $"{Util.hdl(_acfg.Alias, 3)}Idle:False   DropCfg:{dcfg}  TimeCfg:{tcfg} (В очереди на запуск)\n";
                                }
                            }
                            else rzo += $"{Util.hdl(_acfg.Alias, 3)}False: Бот отключен   DropCfg:{dcfg}  TimeCfg:{tcfg}\n";
                        }
                        else rzo += $"False: Бот <{_bot}> не найден\n";
                    }
                }
                if (!String.IsNullOrEmpty(rzo)) SendMsg(Sender, bot, rz+rzo);
                if (!String.IsNullOrEmpty(rzf)) SendMsg(Sender, bot, rz+rzf);
                if (!String.IsNullOrEmpty(rzd)) SendMsg(Sender, bot, rz+rzd);
            }
            else if (_act == "!stop"){
                string[] Aprm = _prm.Split(','); string rz2 = ""; int tsd = 0, tit = 0, tpi = 0;
                try {int.TryParse(Aprm[0], out tsd); int.TryParse(Aprm[1], out tit); int.TryParse(Aprm[2], out tpi);} catch {}
                if (Abots[0] != "all"){
                    foreach (var _bot in Abots) {
                        var _acfg = Worker._accounts.FirstOrDefault(t =>t.Alias==_bot);
                        if (_acfg != null) {
                            if (_acfg.IdleNow) {
                                if ((Acr.Length<2 || Acr[1][0] != '@' || Abots.Length>1) && Worker._accounts.Count(t => t.IdleNow && t.Action != "Stop") < 2) {
                                    rz += $"False: <{_bot}> - это последний активный бот, чтобы его отключить, - укажите его имя: @{_bot}\n";
                                    continue;
                                }
                                if (tpi>0) _acfg.TimeConfig.PauseBeatwinIdleTime = tpi;
                                if (tit>0) {if(tit > MainConfig.Config.ChkIdleTimeOut){_acfg.TimeConfig.IdleTime = tit;} else _acfg.TimeConfig.IdleTime = MainConfig.Config.ChkIdleTimeOut;}
                                if (tsd>0) {_acfg.LastRun = DateTime.UtcNow.AddMinutes(-(_acfg.TimeConfig.IdleTime + _acfg.TimeConfig.PauseBeatwinIdleTime - tsd));}
                                _acfg.Action = "Stop"; rz2 = $"{Util.hdl(_bot, 3)}Готово!\n";
                            }
                            else rz2 = $"False: Бот <{_bot}> уже остановлен\n";
                            if (Aprm[0] == "0" && Aprm.Length == 1){
                                if (_acfg.IdleEnable) {_acfg.IdleEnable = false; rz2 = $"{Util.hdl(_bot, 3)}Готово!\n";}
                                else rz2 = $"False: Бот <{_bot}> уже отключен\n";
                            }
                        }
                        else rz2 = $"False: Бот <{_bot}> не найден\n";
                        rz += rz2;
                    }
                    SendMsg(Sender, bot, rz);
                }
                else {
                    foreach (var accountCfg in Worker._accounts) {
                        if (accountCfg.IdleNow && accountCfg.Alias != bot) {
                            if (tpi>0) accountCfg.TimeConfig.PauseBeatwinIdleTime = tpi;
                            if (tit>0) {if(tit > MainConfig.Config.ChkIdleTimeOut){accountCfg.TimeConfig.IdleTime = tit;} else accountCfg.TimeConfig.IdleTime = MainConfig.Config.ChkIdleTimeOut;}
                            if (tsd>0) {accountCfg.LastRun = DateTime.UtcNow.AddMinutes(-(accountCfg.TimeConfig.IdleTime + accountCfg.TimeConfig.PauseBeatwinIdleTime - tsd));}
                            accountCfg.Action = "Stop";
                        }
                        if (Aprm[0] == "0" && Aprm.Length == 1 && accountCfg.Alias != bot){
                            if (accountCfg.IdleEnable) accountCfg.IdleEnable = false;
                        }
                    }
                    SendMsg(Sender, bot, "/code Готово!");
                }
            }
            else if (_act == "!start"){
                string[] Aprm = _prm.Split(','); int tsd = 0, tit = 0, tpi = 0;
                try {int.TryParse(Aprm[0], out tsd); int.TryParse(Aprm[1], out tit); int.TryParse(Aprm[2], out tpi);} catch {}
                if (Abots[0] != "all" && Abots[0] != "iof" && Abots[0] != "ids"){
                    foreach (var _bot in Abots) {
                        var _acfg = Worker._accounts.FirstOrDefault(t =>t.Alias==_bot);
                        if (_acfg != null) {
                            if (!_acfg.IdleNow) {
                                if (!_acfg.IdleEnable) _acfg.IdleEnable = true;
                                if (tpi>0) _acfg.TimeConfig.PauseBeatwinIdleTime = tpi;
                                if (tit>0) {if(tit > MainConfig.Config.ChkIdleTimeOut){_acfg.TimeConfig.IdleTime = tit;} else _acfg.TimeConfig.IdleTime = MainConfig.Config.ChkIdleTimeOut;}
                                if (tsd>0) {_acfg.LastRun = DateTime.UtcNow.AddMinutes(-(_acfg.TimeConfig.IdleTime + _acfg.TimeConfig.PauseBeatwinIdleTime - tsd));}
                                else _acfg.LastRun = DateTime.MinValue;
                                rz += $"{Util.hdl(_bot, 3)}Готово!\n";
                            }
                            else rz += $"False: Бот <{_bot}> уже запущен\n";
                        }
                        else rz += $"False: Бот <{_bot}> не найден\n";
                    }
                    SendMsg(Sender, bot, rz);
                }
                else {
                    IEnumerable<AccountConfig> _acfgs;
                    if (Abots[0] == "iof"){_acfgs = Worker._accounts.Where(t => !t.IdleNow && t.IdleEnable);}
                    else if (Abots[0] == "ids"){_acfgs = Worker._accounts.Where(t => !t.IdleEnable);}
                    else {_acfgs = Worker._accounts.Where(t => !t.IdleNow);}
                    foreach (var accountCfg in _acfgs) {
                        if (!accountCfg.IdleEnable) accountCfg.IdleEnable = true;
                        if (tpi>0) accountCfg.TimeConfig.PauseBeatwinIdleTime = tpi;
                        if (tit>0) {if(tit > MainConfig.Config.ChkIdleTimeOut){accountCfg.TimeConfig.IdleTime = tit;} else accountCfg.TimeConfig.IdleTime = MainConfig.Config.ChkIdleTimeOut;}
                        if (tsd>0) {accountCfg.LastRun = DateTime.UtcNow.AddMinutes(-(accountCfg.TimeConfig.IdleTime + accountCfg.TimeConfig.PauseBeatwinIdleTime - tsd));}
                        else accountCfg.LastRun = DateTime.MinValue;
                    }
                    SendMsg(Sender, bot, "/code Готово!");
                }
            }
            else if (_act == "!play"){
                string[] Adc = new string[0], Adcf = _prm.Split(';'); var Ldcf = new List<(uint, string)>(); uint It1 = 0;
                try {
                    foreach (var dcf in Adcf) {
                        Adc = dcf.Split(':'); var Adc2 = Adc[1].Split(',');
                        foreach (var dcf2 in Adc2) Int32.Parse(dcf2);
                        if(Adc[0][0] != '-') {It1 = UInt32.Parse(Adc[0]);} else It1 = UInt32.Parse(Adc[0].Substring(1));
                        Ldcf.Add((It1, Adc[1]));
                    }
                }
                catch {SendMsg(Sender, bot, "/code False: Некорректный параметр!"); return;}
                if (Abots[0] != "all" && Abots[0] != "ion" && Abots[0] != "iof" && Abots[0] != "ids"){
                    foreach (var _bot in Abots) {
                        var _acfg = Worker._accounts.FirstOrDefault(t =>t.Alias==_bot);
                        if (_acfg != null) {
                            if (_prm[0] != '-' && _prm[0] != '+') {
                                if(Adc[1]=="-1"){for(var k=0;k<_acfg.DropConfig.Count;k++){if(_acfg.DropConfig[k].Item1==It1 && _acfg.DropConfig[k].Item2[0]!='-')_acfg.DropConfig[k]=(It1,"-1,"+_acfg.DropConfig[k].Item2);}}
                                else if (Adc[1] == "+1"){for(var k=0;k<_acfg.DropConfig.Count;k++){if(_acfg.DropConfig[k].Item1==It1)_acfg.DropConfig[k]=(It1,_acfg.DropConfig[k].Item2.Replace("-1,",""));}}
                                else _acfg.DropConfig = Ldcf.ToList();}
                            else if (_prm[0] == '-') {_acfg.DropConfig.Remove((It1, Adc[1]));}
                            else {_acfg.DropConfig.Remove((It1, Adc[1])); _acfg.DropConfig.Add((It1, Adc[1]));}
                            if (_acfg.IdleNow) _acfg.Action = "!play";
                            rz += $"{Util.hdl(_bot, 3)}Готово!\n";
                        }
                        else rz += $"False: Бот <{_bot}> не найден\n";
                    }
                    SendMsg(Sender, bot, rz);
                }
                else {
                    IEnumerable<AccountConfig> _acfgs;
                    if (Abots[0] == "ion"){_acfgs = Worker._accounts.Where(t => t.IdleNow);}
                    if (Abots[0] == "iof"){_acfgs = Worker._accounts.Where(t => !t.IdleNow && t.IdleEnable);}
                    else if (Abots[0] == "ids"){_acfgs = Worker._accounts.Where(t => !t.IdleEnable);}
                    else {_acfgs = Worker._accounts.Select(t => t);}
                    foreach (var _acf in _acfgs) {
                        if (_prm[0] != '-' && _prm[0] != '+') {
                            if(Adc[1]=="-1"){for(var k=0;k<_acf.DropConfig.Count;k++){if(_acf.DropConfig[k].Item1==It1 && _acf.DropConfig[k].Item2[0]!='-')_acf.DropConfig[k]=(It1,"-1,"+_acf.DropConfig[k].Item2);}}
                            else if (Adc[1] == "+1"){for(var k=0;k<_acf.DropConfig.Count;k++){if(_acf.DropConfig[k].Item1==It1)_acf.DropConfig[k]=(It1,_acf.DropConfig[k].Item2.Replace("-1,",""));}}
                            else _acf.DropConfig = Ldcf.ToList();}
                        else if (_prm[0] == '-') {_acf.DropConfig.Remove((It1, Adc[1]));}
                        else {_acf.DropConfig.Remove((It1, Adc[1])); _acf.DropConfig.Add((It1, Adc[1]));}
                        if (_acf.IdleNow) _acf.Action = "!play";
                    }
                    SendMsg(Sender, bot, "/code Готово!");
                }
            }
            else if (_act == "!drop"){
                string[] Aprm = _prm.Split(','); string rz2 = ""; int dln;
                foreach (var _bot in Abots) {
                    if (!int.TryParse(Aprm[0], out dln)) dln = 10;
                    var _acfg = Worker._accounts.FirstOrDefault(t =>t.Alias==_bot);
                    if (_acfg != null) {
                        rz2 = rz + $"{Util.hdl(_bot, 3)}История дропа:\n";
                        if (File.Exists(Path.Combine(MainConfig.Config.DropHistoryFolder, _acfg.Name+".txt"))) {
                            string[] drops = File.ReadAllLines(Path.Combine(MainConfig.Config.DropHistoryFolder, _acfg.Name+".txt"));
                            if (drops.Length > dln) {dln = drops.Length - dln;} else dln = 0;
                            if (Aprm.Length < 2) {for(int i = dln; i < drops.Length; i++) rz2 += drops[i].Replace("'", "") + "\n";}
                            else for(int i = dln; i < drops.Length; i++) if (drops[i].Contains($": {Aprm[1]} -")) rz2 += drops[i].Replace("'", "") + "\n";
                        } //else Console.WriteLine($"Файл не нейден:{_acfg.Name}.txt");
                    }
                    else rz2 = rz + $"False: Бот <{_bot}> не найден\n";
                    SendMsg(Sender, bot, rz2);
                }
            }
            else if (_act == "!online"){
                if (_prm != "on" && _prm != "off") {SendMsg(Sender, bot, "/code False: Некорректный параметр!"); return;}
                foreach (var _bot in Abots) {
                    var _acfg = Worker._accounts.FirstOrDefault(t =>t.Alias==_bot);
                    if (_acfg != null) {
                        if (_prm == "on") {
                          if (_acfg.IdleNow) _acfg.Action = "!chst";
                          _acfg.ShowStatus = "Online";
                        } else {
                          if (_acfg.IdleNow) _acfg.Action = "!chst";
                          _acfg.ShowStatus = "Offline";
                        }
                        rz += $"{Util.hdl(_bot, 3)}Готово!\n";
                    }
                    else rz += $"False: Бот <{_bot}> не найден\n";
                }
                SendMsg(Sender, bot, rz);
            }
            else if (_act == "!ren"){
                var _bot = Abots[0];
                var _acfg = Worker._accounts.FirstOrDefault(t =>t.Alias==_bot);
                if (_acfg != null) {
                    if (!_acfg.IdleEnable) {
                        string newnm = Regex.Replace(_prm, @"[^0-9A-Za-z]", "");
                        int naml = newnm.Length;
                        if (naml > 0) {
                            if (!(Worker._accounts.Count(x => x.Alias == newnm) > 0)) {
                                if (naml > Util.unmax) Util.unmax = naml;
                                _acfg.Alias = newnm; _acfg.Save();
                                rz += $"Бот <{_bot}> переименован в <{newnm}>!\n";
                            } else rz += $"False: Бот <{newnm}> уже существует!\n";
                        } else rz += $"False: некорректное имя!\n";
                    } else rz += $"False: бот <{_bot}> должен быть отключенным!\n";
                }
                else rz += $"False: Бот <{_bot}> не найден\n";
                SendMsg(Sender, bot, rz);
            }
            else if (_act == "!save"){
                foreach (var _bot in Abots) {
                    var _acfg = Worker._accounts.FirstOrDefault(t =>t.Alias==_bot);
                    if (_acfg != null) {
                        _acfg.Save();
                        rz += $"{Util.hdl(_bot, 3)}Готово!\n";
                    }
                    else rz += $"False: Бот <{_bot}> не найден\n";
                }
                SendMsg(Sender, bot, rz);
            }
            else if (_act == "!help"){
                string rzh = "/code Список команд:\n";
                rzh += "!stat @bot1		Выводит информацию о статусе бота. @ - символ, обозначающий, что bot1 это имя бота. Можно указывать несколько ботов: @bot1,bot2 Также есть несколько спец. имен: ";
                rzh += "@all (все боты), @ion (активные боты), @iof (остановленные боты), @ids (отключенные боты). Можно и не указывать имя бота, в таком случае команда будет адресована боту, который был";
                rzh += " выбран в чате.\n\n";
                rzh += "!start @bot1 5,100,20	Запуск бота/ботов. 5 - первый параметр устанавливает задержку запуска бота в минутах (0 - немедленный запуск). 100,20 - второй и третий параметр позволяют ";
                rzh += "изменять значения IdleTime (время фарма в минутах) и PauseBeatwinIdleTime (время простоя до следующего фарма) в конфигах ботов. Все три параметра являются необязательными, и если ";
                rzh += "они отсутствуют или равны 0, то соответствующее значение не будет изменено.\n\n";
                rzh += "!stop @bot1 5,100,20	Остановка или отключение бота/ботов. 5 - первый параметр устанавливает задержку запуска бота в минутах после его остановки (0 без 2-го и 3-го параметра ";
                rzh += "ОТКЛЮЧАЕТ бота). 100,20 - второй и третий параметр позволяют изменять значения IdleTime и PauseBeatwinIdleTime в конфигах ботов. Все три параметра являются необязательными.\n\n";
                rzh += "!play @bot1 218620:1;232090:900000,910000	Управление списком фармящихся игр (изменяет значения DropConfig). 218620(id 1-й игры) 1(id_дропа 1-й игры); 232090(id 2-й игры) ";
                rzh += "900000,910000(id_дроп1,id_дроп2 2-й игры). Можно также использовать спец. параметры: 218620:-1 (временно отключить фарм указанной игры (+1 включить обратно)). -218620:1 убрать из ";
                rzh += "DropConfig игру с соответствующим app_id и drop_id. +218620:1 добавить игру в список DropConfig.\n\n";
                rzh += "!drop @bot1 5,218620	Выводит историю дропа бота/ботов. 5 - первый параметр указывает количество строк дропа (по умолчанию 10 строк), 218620 - id игры, по которой будет отфильтрован";
                rzh += " список дропа (необязательный параметр).\n\n";
                rzh += "!online @bot1 on	Изменение статуса бота/ботов в сети: on/off.\n\n";
                rzh += "!ren @bot1 newname	Переименование бота (изменение имени (Alias) бота).\n\n";
                rzh += "!save @bot1		Сохранение конфига бота/ботов. Автоматическое сохранение конфигов ботов происходит только при их запуске и остановке. Поэтому если в конфигах были сделаны изменения ";
                rzh += "и вы хотите их сохранить, - используйте команду !save\n\n";
                rzh += "!exit 	Завершение работы дроплера.";
                SendMsg(Sender, bot, rzh);
            }
            else if (_act == "!exit"){
                _steamAccount.Action = "!exit";
                SendMsg(Sender, bot, "Готово!");
            }
            else {
                SendMsg(Sender, bot, "/code False: Неизвестная команда! (Наберите !help для справки)");
            }
        }

        void SendMsg(ulong Sender, string bot, string message) {
            CFriendMessages_SendMessage_Request request = new CFriendMessages_SendMessage_Request {
                chat_entry_type = (int) EChatEntryType.ChatMsg,
                contains_bbcode = true,
                message = message,
                steamid = Sender
            };
            try {
                UnifiedFriendMessagesService.SendMessage(x => x.SendMessage(request));
            } catch (Exception e) {
                Console.ForegroundColor = ConsoleColor.Red;
                Console.WriteLine($"{Util.hdl(bot)}{e}");
                Console.ResetColor();
            }
        }
        
    }
}
